#!/usr/bin/env python
#
# Data Quality Checker also known as DQC
#
# Read the README.md for more detailed information
#
#       Barcelona Supercomputing Center - Centro Nacional de Supercomputacion
#       Earth Sciences Department - Computational Earth Sciences
#     
#     Authors: 
#       Joan Sala Calero (joan.sala@bsc.es)
#       Jesus Pena-Izquierdo (jesus.pena@bsc.es)

import sys

import warnings
warnings.filterwarnings('ignore')

# Local imports
from dqc_wrapper.utils.utils_forms import  *
from dqc_wrapper.utils.utils_xarray import *
from dqc_wrapper.utils.utils_bokeh import *
from dqc_wrapper.utils.utils_files import *
from dqc_wrapper.utils.utils_grib import *
from dqc_wrapper.utils.utils_stats import *

"""
Class: Valid ranges checks
---
- Generate histograms of data values per month
- If valid_min/valid_max are provided, raise error if outlier found
"""
class ValidRangesTest:

    def __init__(self, resdir, testname, files, varname, vmin, vmax, dt, forms_dir, cds_dataset, cds_variable):
        self.testid = 'test06-valid_ranges'
        self.logger = logging.getLogger(self.testid)
        self.output_plot_html = os.path.join(resdir, testname.replace('.ini', '-valid_ranges.html'))
        self.csv_hist = os.path.join(resdir, testname.replace('.ini', '-hist.csv'))
        self.csv_stats = os.path.join(resdir, testname.replace('.ini', '-stats.csv'))
        self.files = files
        self.varname = varname     
        self.datatype = dt
        self.stat_columns = [ 'Min', 'Max', 'Med', 'q1', 'q3', 'IQR' ]
        self.cds_variable = cds_variable
        self.cds_dataset = cds_dataset
        self.metadata = get_metadata(forms_dir, cds_dataset, cds_variable)

        # Automatic versus computed valid ranges
        try:
            self.valid_min = float(vmin)
        except:
            self.valid_min = None
        try:
            self.valid_max = float(vmax)
        except:
            self.valid_max = None

    """
    Loading dataset | All Timesteps+Datatypes 
    Tries to load data, return exception message otherwise
    """
    def load_datafile(self, filepath):
        # Read data into Xarray
        try:
            ds = open_dataset(filepath, datatypes=self.datatype)
        except Exception as e:
            return None, str(e)

        return ds, ''

    """
    Load all data values from a given dataset list (product types) - Avoid NaN
    """
    def load_arrays(self, ds):
        # Load data from all product types into one array
        arr = []
        for d in ds:
            # Concat values (avoid NaN)
            tmp_arr = get_variable(d, self.varname).flatten()
            tmp_arr = tmp_arr[~np.isnan(tmp_arr)]
            arr = np.append(arr, tmp_arr)

        return arr

    """
    Hadley cell stats (90, 60, 30) absolute latitudes (polar, mid-latitude, hadley) - Remove NaN 
    """
    def get_hist_hadley_xarray(self, df_hist, hbins, ds, init, lat0, lat1, row, latid):
        arr = ds[self.varname].where((ds[latid] > lat0) & (ds[latid] <= lat1)).values
        arr = arr[~np.isnan(arr)]
        x, b = np.histogram(arr, bins=hbins)

        # Initialization vs sum
        if init==0:
            df_hist.loc[row] = x
        else:
            df_hist.loc[row] += x

        return x, [ arr.min(), arr.max() ]

    """
    Get Histogram bins initalization. 
    Number of bins + K = number of times (amplification of the histogram bins) 
    Initialization observed minmax
    """
    def get_bins_init(self, files, nbins=2000, Kmax=5):

        # XARRAY (NetCDF)
        if files[0].endswith('.nc'):
            # Try to load data from the first file
            ds, errmsg = self.load_datafile(files[0])
            if errmsg != '':
                return None, None, None, errmsg

            # Load all data from all product types
            arr = self.load_arrays(ds)

        # ECCODES (GRIB)
        else:
            # Get a sample from every first message of every file [better, faster, stronger]
            arr = []
            for f in files:
                gf = GribFileCheck(f)
                arr += gf.get_grib_first_values().tolist() # Get sample (of file X)

        # Return histogram initialization
        minv = np.min(arr)
        maxv = np.max(arr)
        dist = abs(maxv-minv)
        return np.histogram_bin_edges(arr, bins=nbins, range=(minv-Kmax*dist, maxv+Kmax*dist)), minv, maxv, ''

    """
    Main function
    """
    def check_valid_ranges(self):
       
        # Progress init
        i = 0
        n = len(self.files) # read/calc_stats

        # Store stats information per file (basename is the unique key and contains date)
        bn = [ os.path.basename(f) for f in self.files ] 
        df_stats = pd.DataFrame(index=bn, columns=self.stat_columns)

        # Init histogram bins with the first file
        hist_bins, minv, maxv, errmsg = self.get_bins_init(self.files)
        if errmsg != '':
            return 'err', errmsg

        # We keep statistics per hadley-cell, filename and total [order is relevant]
        hadley_cells = {
            'N_polar': [60, 90],
            'N_midlat': [30, 60],
            'N_hadley': [0, 30],
            'S_hadley': [-30, 0],
            'S_midlat': [-60, -30],
            'S_polar': [-90, -60]
        }
        df_hist = pd.DataFrame(
            index = bn + list(hadley_cells.keys()),
            columns = hist_bins[:-1]
        )
       
        # Generate histograms, store them in a dataframe
        ind = 0
        for f in self.files:
            # Progress update
            i+=1
            print_progress(i, n)              

            # NetCDF - read all datatypes individually
            if f.endswith('.nc'):
                # Load xarray dataset
                ds, errmsg = self.load_datafile(f)
                if errmsg != '':
                    return 'err', errmsg

                # Latitude histograms - subset and cumulative sum for 6-hadley cells
                for d in ds:
                    latid = get_lat_name(d)
                    first = True

                    for hc, lats in hadley_cells.items():
                        hc_hist, mnmx = self.get_hist_hadley_xarray(
                            df_hist, hist_bins, d, ind,  lats[0],  lats[1], hc, latid
                        )

                        # Compute histogram (global/month) as sum of all latitude ranges (init vs accum)
                        if first:
                            xsum = hc_hist
                            minv = min(mnmx)
                            maxv = max(mnmx)
                        else:
                            xsum = sum([hc_hist, xsum])
                            minv = min(minv, mnmx)
                            maxv = max(maxv, mnmx)
            else:
                # All regions at once (more efficient) - message x message
                ds = GribFileCheck(f)
                xsum, minv, maxv = ds.get_grib_hadley_regions(ind, hist_bins, df_hist)

            # Store general/full histogram
            df_hist.loc[os.path.basename(f)] = xsum

            # Compute stats (of the histogram - faster)
            df_stats.loc[os.path.basename(f)] = [ minv, maxv ] + get_basic_stats(xsum, hist_bins)
            
            # Next iteration
            del ds
            del xsum
            ind+=1

        # Compute total histogram        
        df_hist.loc['total'] = df_hist.sum().values
        
        # Save stats for future use
        df_hist.index.name = 'File'
        df_hist.to_csv(self.csv_hist)
        print('INFO: Saving histogram to {}'.format(self.csv_hist))
        df_stats.index.name = 'File'
        df_stats.to_csv(self.csv_stats)
        print('INFO: Saving stats to {}'.format(self.csv_stats))

        # Defaults
        res, msg = '', ''

        # Just histogram delivered, unknown ranges for the variable
        if self.valid_min is None and self.valid_max is None:
            res = 'warn'
            msg = 'Physical plausible range unknown. Data range: [ from {} to  {} ] {}'.format(
                format_float(min(df_stats['Min'])), format_float(max(df_stats['Max'])), self.metadata['unit']
            )

        # Valid range estimation (only if defined, well-known variable limits)               
        else:
            for fname, row in df_stats.iterrows():
                # Check ranges          
                if row['Min'] < self.valid_min or row['Max'] > self.valid_max:
                    self.logger.error('{} contains outliers'.format(fname))
                    self.logger.error('[ {} > {} ] vs [ {} > {} ]'.format(self.valid_min, self.valid_max, row['Min'], row['Max']))
                    res = 'err'               

            # Conclusion plausible range            
            if res == '':
                res = 'ok'
                msg = 'No outliers found. Plausible range: [ from {} to  {} ]'.format(self.valid_min, self.valid_max)      
            else:
                msg = 'Potential outliers found. Plausible range: [ from {} to  {} ]'.format(self.valid_min, self.valid_max)

        # Final plots
        bp = BokehPlot(self.output_plot_html, self.csv_stats, self.csv_hist, self.cds_variable,
                       self.valid_min, self.valid_max, self.metadata)
        bp.plot_histograms()
                
        # Final message
        self.logger.info(msg)

        return res, msg

    def main(self):
       
        # Do the actual work
        res, msg = self.check_valid_ranges()

        # Prepare output
        return { 
            'res': res, 
            'msg': msg,
            'plot': self.output_plot_html
        }    
    
if __name__ == '__main__':

    resdir = '/tmp/test_valid_ranges'
    testname = 'test_valid_ranges.ini'
    files = [ '/data/dqc_test_data/reanalysis/reanalysis-era5-single-levels/sea_surface_temperature/reanalysis-20051219.grib',
              '/data/dqc_test_data/reanalysis/reanalysis-era5-single-levels/sea_surface_temperature/ensemble_mean-20051219.grib' ]

    vmin = ''
    vmax = ''
    forms_dir = '/data/cds-forms-c3s'
    cds_dataset = 'reanalysis-era5-single-levels'
    cds_variable = 'sea_surface_temperature'
    varname = 'sst'
    dt = 'an'

    vr = ValidRangesTest(resdir, testname, files, varname, vmin, vmax, dt, forms_dir, cds_dataset, cds_variable)
    print(vr.check_valid_ranges())

    sys.exit()
